

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>

#include <unistd.h>


#include <Idx_private.h>
#include <Idx/Idx.h>

//#define inline


//################################################################################
//###
//###   Scanning/printing utilities
//###


#define _IDX_INTSIZE 10
#define _IDX_FLPSIZE 15
#define _IDX_FLPFORM "0.9"


unsigned long _Idx_printint(unsigned long u, char * s, unsigned long m) {
	register unsigned long i = 0 ;
	if(s==NULL) return 10 ;
	if(m>10) m = 10 ;
	i = m ;
	while(i>0) {
		i -- ;
		s[i] = (u%10)+'0' ;
		u /= 10 ;
	}
	return m ;
}


unsigned long _Idx_scanint(unsigned long * u, char * s, unsigned long m) {
	register unsigned long i = 0 ;
	char c ;
	if(u==NULL) return 0 ;
	if(s==NULL) return 0 ;
	(*u) = 0 ;
	while(i<m) {
		if(s[i]=='\0') break ;
		if( (s[i]>='0') && (s[i]<='9') ) c = s[i]-'0' ;
		else break ;
		(*u) *= 10 ;
		(*u) += c ;
		i ++ ;
	}
	return i ;
}


//################################################################################
//###
//###   DataBase handle
//###

/*
struct S_Idx_DB {
	char * name ;
	int filedes ;
	unsigned long id ;
	unsigned long size ;
	unsigned long wlen ;
	unsigned long flags ;
	unsigned long data ;
} ;

typedef struct S_Idx_DB * Idx_DB ;
*/

#define _IDX_MAGICWLEN 12
#define _IDX_MAGICNLEN 4
#define _IDX_HEADERLEN (_IDX_MAGICWLEN+_IDX_MAGICNLEN+1+5*(_IDX_INTSIZE+1))

// A Magic sequence to check for file format...
char _Idx_Sym[_IDX_MAGICWLEN+1] = "IdxDataBase-" ;
char _Idx_Mag[_IDX_MAGICNLEN+1] = "E%Yc" ;


Idx_DB _Idx_DBRead(Idx_DB db) {
	char buf[_IDX_HEADERLEN] ;
	unsigned long i, r ;
	if(db==NULL) {
		return NULL ;
	}
	if(db->filedes==-1) {
		return NULL ;
	}
	r = read(db->filedes,buf,_IDX_HEADERLEN) ;
	if(r==-1) {
		return NULL ;
	}
	if(r<_IDX_HEADERLEN) {
		return NULL ;
	}
	buf[r] = 0 ;
	// Syntax check
	i = 0 ;
	// Checking magic sequence
	r = 0 ;
	while(r<_IDX_MAGICWLEN) {
		if(buf[i]!=_Idx_Sym[r]) {
			return NULL ;
		}
		r ++ ;
		i ++ ;
	}
	r = 0 ;
	while(r<_IDX_MAGICNLEN) {
		if(buf[i]!=_Idx_Mag[r]) {
			return NULL ;
		}
		r ++ ;
		i ++ ;
	}
	if(buf[i]!=' ') {
		return NULL ;
	}
	i ++ ;
	// Checking ID
	if(_Idx_scanint(&(db->id),buf+i,_IDX_INTSIZE)<_IDX_INTSIZE) {
		return NULL ;
	}
	i += _IDX_INTSIZE ;
	if(buf[i]!=' ') {
		return NULL ;
	}
	i ++ ;
	// Checking size
	if(_Idx_scanint(&(db->size),buf+i,_IDX_INTSIZE)<_IDX_INTSIZE) {
		return NULL ;
	}
	i += _IDX_INTSIZE ;
	if(buf[i]!=' ') {
		return NULL ;
	}
	i ++ ;
	// Checking wlen
	if(_Idx_scanint(&(db->wlen),buf+i,_IDX_INTSIZE)<_IDX_INTSIZE) {
		return NULL ;
	}
	i += _IDX_INTSIZE ;
	if(buf[i]!=' ') {
		return NULL ;
	}
	i ++ ;
	// Checking flen
	if(_Idx_scanint(&(db->flen),buf+i,_IDX_INTSIZE)<_IDX_INTSIZE) {
		return NULL ;
	}
	i += _IDX_INTSIZE ;
	if(buf[i]!=' ') {
		return NULL ;
	}
	i ++ ;
	// Checking vlen
	if(_Idx_scanint(&(db->vlen),buf+i,_IDX_INTSIZE)<_IDX_INTSIZE) {
		return NULL ;
	}
	i += _IDX_INTSIZE ;
	if(buf[i]!='\n') {
		return NULL ;
	}
	// Checking record data length
	r = lseek(db->filedes,_IDX_HEADERLEN+db->size*(db->wlen+_IDX_INTSIZE+2),SEEK_SET) ;
	if(r==-1) {
		return NULL ;
	}
	r = read(db->filedes,buf,_IDX_INTSIZE+1) ;
	if(r==-1) {
		return NULL ;
	}
	if(r<(_IDX_INTSIZE+1)) { // Carry return needed!
		return NULL ;
	}
	r = _Idx_scanint(&(db->data),buf,_IDX_INTSIZE) ;
	if(r<_IDX_INTSIZE) {
		return NULL ;
	}
	// Override any data length when table is empty
	if(db->size==0) db->data = 0 ;
	return db ;
}


Idx_DB _Idx_DBWrite(Idx_DB db) {
	char buf[_IDX_HEADERLEN] ;
	unsigned long i, r ;
	if(db==NULL) {
		return NULL ;
	}
	if(db->filedes==-1) {
		return NULL ;
	}
	// Updating the header line
	i = 0 ;
	// Copying magic sequence
	r = 0 ;
	while(r<_IDX_MAGICWLEN) {
		buf[i] = _Idx_Sym[r] ;
		r ++ ;
		i ++ ;
	}
	r = 0 ;
	while(r<_IDX_MAGICNLEN) {
		buf[i] = _Idx_Mag[r] ;
		r ++ ;
		i ++ ;
	}
	buf[i] = ' ' ;
	i ++ ;
	// Copying ID
	r = _Idx_printint(db->id,buf+i,_IDX_INTSIZE) ;
	if(r<_IDX_INTSIZE) {
		errno = EIO ;
		return NULL ;
	}
	i += _IDX_INTSIZE ;
	buf[i] = ' ' ;
	i ++ ;
	// Copying size
	r = _Idx_printint(db->size,buf+i,_IDX_INTSIZE) ;
	if(r<_IDX_INTSIZE) {
		errno = EIO ;
		return NULL ;
	}
	i += _IDX_INTSIZE ;
	buf[i] = ' ' ;
	i ++ ;
	// Copying wlen
	r = _Idx_printint(db->wlen,buf+i,_IDX_INTSIZE) ;
	if(r<_IDX_INTSIZE) {
		errno = EIO ;
		return NULL ;
	}
	i += _IDX_INTSIZE ;
	buf[i] = ' ' ;
	i ++ ;
	// Copying flen
	r = _Idx_printint(db->flen,buf+i,_IDX_INTSIZE) ;
	if(r<_IDX_INTSIZE) {
		errno = EIO ;
		return NULL ;
	}
	i += _IDX_INTSIZE ;
	buf[i] = ' ' ;
	i ++ ;
	// Copying vlen
	r = _Idx_printint(db->vlen,buf+i,_IDX_INTSIZE) ;
	if(r<_IDX_INTSIZE) {
		errno = EIO ;
		return NULL ;
	}
	i += _IDX_INTSIZE ;
	buf[i] = '\n' ;
	i ++ ;
	// Writing the header line
	r = lseek(db->filedes,0,SEEK_SET) ;
	if(r==-1) {
		return NULL ;
	}
	r = write(db->filedes,buf,_IDX_HEADERLEN) ;
	if( (r==-1)||(r<_IDX_HEADERLEN) ) {
		return NULL ;
	}

	// Updating the record data length
	// Copying data length
	r = _Idx_printint(db->data,buf,_IDX_INTSIZE) ;
	if(r<_IDX_INTSIZE) {
		errno = EIO ;
		return NULL ;
	}
	buf[_IDX_INTSIZE] = '\n' ;
	// Writing data length
	r = lseek(db->filedes,_IDX_HEADERLEN+db->size*(db->wlen+_IDX_INTSIZE+2),SEEK_SET) ;
	if(r==-1) {
		return NULL ;
	}
	r = write(db->filedes,buf,_IDX_INTSIZE+1) ;
	if( (r==-1)||(r<(_IDX_INTSIZE+1)) ) {
		return NULL ;
	}
	return db ;
}


Idx_DB Idx_DBCreate(char * file, unsigned long id, unsigned long wlen, unsigned long flen, unsigned long vlen, unsigned long flags) {
	int fd ;
	Idx_DB db ;
	unsigned long l ;
	unsigned long i ;
	if(file==NULL) {
		errno = EINVAL ;
		return NULL ;
	}
	fd = open(file,O_RDWR|O_CREAT|O_EXCL,0644) ;
	if(fd==-1) {
		return NULL ;
	}
	db = malloc(sizeof(struct S_Idx_DB)) ;
	if(db==NULL) {
		close(fd) ;
		return NULL ;
	}
	db->flags = flags ;
	l = 0 ;
	while(file[l]!=0) l ++ ;
	db->name = malloc(l+1) ;
	if(db->name==NULL) {
		close(fd) ;
		free(db) ;
		return NULL ;
	}
	i = 0 ;
	while(i<l) {
		db->name[i] = file[i] ;
		i ++ ;
	}
	db->name[i] = 0 ;
	db->filedes = fd ;
	db->size = 0 ;
	db->wlen = wlen ;
	db->flen = flen ;
	db->vlen = vlen ;
	db->data = 0 ;
	if(_Idx_DBWrite(db)==NULL) {
		return NULL ;
	}
	return db ;
}


Idx_DB Idx_DBOpen(char * file, unsigned long flags) {
	int fd ;
	Idx_DB db ;
	unsigned long l ;
	unsigned long i ;
	if(file==NULL) {
		errno = EINVAL ;
		return NULL ;
	}
	fd = open(file,O_RDWR) ;
	if(fd==-1) {
		return NULL ;
	}
	db = malloc(sizeof(struct S_Idx_DB)) ;
	if(db==NULL) {
		close(fd) ;
		return NULL ;
	}
	db->flags = flags ;
	l = 0 ;
	while(file[l]!=0) l ++ ;
	db->name = malloc(l+1) ;
	if(db->name==NULL) {
		close(fd) ;
		free(db) ;
		return NULL ;
	}
	i = 0 ;
	while(i<l) {
		db->name[i] = file[i] ;
		i ++ ;
	}
	db->name[i] = 0 ;
	db->filedes = fd ;
	if(_Idx_DBRead(db)==NULL) {
		errno = EIO ;
		return NULL ;
	}
	return db ;
}


void Idx_DBClose(Idx_DB db) {
	if(db!=NULL) {
		_Idx_DBWrite(db) ;
		close(db->filedes) ;
		db->filedes = -1 ;
		free(db->name) ;
		db->name = NULL ;
		db->id = 0 ;
		db->size = 0 ;
		db->wlen = 0 ;
	}
}

void Idx_DBStat(Idx_DB db) {
	if(db==NULL) {
		printf("%s%s: *\n",_Idx_Sym,_Idx_Mag) ;
		return ;
	}
	printf("%s%s: %10.10lu\n",_Idx_Sym,_Idx_Mag,db->id) ;
	printf("\tName:     \"%s\"\n",db->name) ;
	printf("\tSize:     %lu\n",db->size) ;
	printf("\tWordLen:  %lu\n",db->wlen) ;
	printf("\tFieldLen: %lu\n",db->flen) ;
	printf("\tValueLen: %lu\n",db->vlen) ;
	printf("\tData seg: %lu\n",db->data) ;
}



//################################################################################
//###
//###   Index retrival and querying
//###

/*
struct S_Idx_Index {
	Idx_DB db ;
	unsigned long pos ;
	char * word
	double eval ;
	unsigned long off ;
	unsigned long mem ;
	unsigned long len ;
	double eval ;
	char * * flds ;
	char * * vals ;
} ;

typedef struct S_Idx_Index * Idx_Index ;
*/

#define _IDX_INDEX_BSIZE 8


////////////////////////////////////////
// Index object management


void Idx_IndexFree(Idx_Index idx) {
	if(idx!=NULL) {
		if(idx->flds!=NULL) {
			unsigned long i = 0 ;
			while(i<idx->len) {
				free(idx->flds[i]) ;
				i++ ;
			}
		}
		if(idx->vals!=NULL) {
			unsigned long i = 0 ;
			while(i<idx->len) {
				free(idx->vals[i]) ;
				i++ ;
			}
		}
		free(idx->vals) ;
		idx->vals = NULL ;
		free(idx->flds) ;
		idx->flds = NULL ;
		free(idx->word) ;
		idx->word = NULL ;
		idx->db = NULL ;
		free(idx) ;
	}
}


void Idx_IndexStat(Idx_Index idx) {
	if(idx==NULL) {
		printf("Idx_Index: *\n") ;
		return ;
	}
	printf("Idx_Index: %10.10lu \"%s\"\n",idx->pos,idx->word) ;
	if(idx->db==NULL) {
		printf("\tIdx_DB: *\n") ;
	}
	else {
		printf("\tIdxDB: %10.10lu \"%s\"\n",idx->db->id,idx->db->name) ;
	}
	printf("\tOff:   %lu\n",idx->off) ;
	printf("\tMem:   %lu\n",idx->mem) ;
	printf("\tLen:   %lu\n",idx->len) ;
	printf("\tEval:  %f\n",idx->eval) ;
}


void Idx_IndexPrint(Idx_Index idx) {
	unsigned long i = 0 ;
	if(idx==NULL) {
		printf("(NIL)\n") ;
		return ;
	}
	printf("<%s> %f\n",idx->word,idx->eval) ;
	while(i<idx->len) {
		printf("  %*s => %s\n",-(int)idx->db->flen,idx->flds[i],idx->vals[i]) ;
		i ++ ;
	}
}



////////////////////////////////////////
// Index retrival / storage



Idx_Index Idx_IndexAlloc(Idx_DB db, char * word) {
	unsigned long i = 0 ;
	Idx_Index idx ;
	if(db==NULL) {
		errno = EINVAL ;
		return NULL ;
	}
	idx = malloc(sizeof(struct S_Idx_Index)) ;
	if(idx==NULL) {
		return NULL ;
	}
	idx->db = db ;
	idx->word = malloc(db->wlen*sizeof(char)) ;
	if(idx->word==NULL) {
		free(idx) ;
		return NULL ;
	}
	while(i<db->wlen) {
		idx->word[i] = word[i] ;
		if(word[i]==0) break ;
		i ++ ;
	}
	while(i<db->wlen) {
		idx->word[i] = 0 ;
		i ++ ;
	}
	idx->pos = -1 ;
	idx->off = -1 ;
	idx->eval = 0 ;
	idx->mem = 0 ;
	idx->len = 0 ;
	idx->flds = NULL ;
	idx->vals = NULL ;
	return idx ;
}


Idx_Index _Idx_IndexRead(Idx_DB db, unsigned long pos) {
	Idx_Index idx = NULL ;
	long r ;
	char buf[_IDX_INTSIZE+_IDX_FLPSIZE+2] ;
	char * rbuf ;
	unsigned long o = 0;
	char * word = NULL ;
	unsigned long l = 0;
	if(db==NULL) {
		return NULL ;
	}
	if(pos>=db->size) {
		return NULL ;
	}
	// Reading offset
	r = lseek(db->filedes,_IDX_HEADERLEN+pos*(db->wlen+_IDX_INTSIZE+2),SEEK_SET) ;
	if(r==-1) {
		return NULL ;
	}
	r = read(db->filedes,buf,_IDX_INTSIZE+1) ;
	if((r==-1)||(r<(_IDX_INTSIZE+1))||(buf[_IDX_INTSIZE]!=' ')) {
		return NULL ;
	}
	if(_Idx_scanint(&o,buf,_IDX_INTSIZE)<_IDX_INTSIZE) {
		return NULL ;
	}
	// Reading word
	word = malloc((db->wlen+1)*sizeof(char)) ;
	if(word==NULL) {
		return NULL ;
	}
	r = read(db->filedes,word,db->wlen+1) ;
	if((r==-1)||(r<(db->wlen+1))||(word[db->wlen]!='\n')) {
		free(word) ;
		return NULL ;
	}
	word[db->wlen] = 0 ;
	// Reading record length
	r = lseek(db->filedes,o+_IDX_HEADERLEN+_IDX_INTSIZE+1+db->size*(db->wlen+_IDX_INTSIZE+2),SEEK_SET) ;
	if(r==-1) {
		return NULL ;
	}
	r = read(db->filedes,buf,_IDX_INTSIZE+1) ;
	if((r==-1)||(r<(_IDX_INTSIZE+1))||(buf[_IDX_INTSIZE]!=' ')) {
		return NULL ;
	}
	if(_Idx_scanint(&l,buf,_IDX_INTSIZE)<_IDX_INTSIZE) {
		return NULL ;
	}
	idx = malloc(sizeof(struct S_Idx_Index)) ;
	if(idx==NULL) {
		return NULL ;
	}
	idx->db = db ;
	idx->pos = pos ;
	idx->word = word ;
	idx->off = o ;
	idx->mem = _IDX_INDEX_BSIZE*(l/_IDX_INDEX_BSIZE+1) ;
	idx->len = l ;
	idx->eval = 0 ;
	idx->flds = NULL ;
	idx->vals = NULL ;
	// Reading value
	r = read(db->filedes,buf,_IDX_FLPSIZE+1) ;
	if((r==-1)||(r<(_IDX_FLPSIZE+1))||((l==0)&&(buf[_IDX_FLPSIZE]!='\n'))||((l>0)&&(buf[_IDX_FLPSIZE]!=' '))) {
		return NULL ;
	}
	buf[_IDX_FLPSIZE] = 0 ;
	if(sscanf(buf,"%lf",&(idx->eval))<1) {
		idx->eval = 0 ;
	}
	// Record members loop
	idx->flds = malloc(idx->mem*sizeof(char *)) ;
	if(idx->flds==NULL) {
		free(idx) ;
		return NULL ;
	}
	idx->vals = malloc(idx->mem*sizeof(char *)) ;
	if(idx->vals==NULL) {
		free(idx->flds) ;
		free(idx) ;
		return NULL ;
	}
	rbuf = alloca(db->flen+db->vlen+2) ;
	if(rbuf==NULL) {
		free(idx->flds) ;
		free(idx->vals) ;
		free(idx) ;
		return NULL ;
	}
	l = 0 ;
	while(l<idx->len) {
		unsigned long i = 0 ;
		unsigned long j ;
		idx->flds[l] = malloc((db->flen+1)*sizeof(char)) ;
		if(idx->flds[l]==NULL) break ;
		idx->vals[l] = malloc((db->vlen+1)*sizeof(char)) ;
		if(idx->vals[l]==NULL) { free(idx->flds[l]) ; break ; }
		r = read(db->filedes,rbuf,db->flen+db->vlen+2) ;
		if( (r==-1) || (r<(db->flen+db->vlen+2))
			|| (rbuf[db->flen]!=':')
			|| ((rbuf[db->flen+db->vlen+1]!=' ')&&(rbuf[db->flen+db->vlen+1]!='\n'))
			) {
			free(idx->flds[l]) ;
			free(idx->vals[l]) ;
			break ;
		}
		j = 0 ;
		while(j<db->flen) {
			idx->flds[l][j] = rbuf[i] ;
			i ++ ;
			j ++ ;
		}
		idx->flds[l][j] = 0 ;
		i ++ ; // jump the ':'
		j = 0 ;
		while(j<db->vlen) {
			idx->vals[l][j] = rbuf[i] ;
			i ++ ;
			j ++ ;
		}
		idx->vals[l][j] = 0 ;
		l ++ ;
	}
	idx->len = l ; // Makes sure not found items are not counted
	return idx ;
}



Idx_Index _Idx_IndexWrite(Idx_DB db, Idx_Index idx, unsigned long pos) {
	char buf[_IDX_INTSIZE+_IDX_FLPSIZE+2] ;
	char * rbuf ;
	unsigned long i, r ;
	if(db==NULL) {
		return NULL ;
	}
	if(idx==NULL) {
		return NULL ;
	}
	if(idx->off==-1) {
		return NULL ;
	}
	rbuf = alloca(db->flen+db->vlen+2) ;
	if(rbuf==NULL) {
		return NULL ;
	}
	// Writing the record offset
	r = _Idx_printint(idx->off,buf,_IDX_INTSIZE) ;
	if(r<_IDX_INTSIZE) {
		return NULL ;
	}
	buf[_IDX_INTSIZE] = ' ' ;
	r = lseek(db->filedes,_IDX_HEADERLEN+pos*(db->wlen+_IDX_INTSIZE+2),SEEK_SET) ;
	if(r==-1) {
		return NULL ;
	}
	r = write(db->filedes,buf,_IDX_INTSIZE+1) ;
	if(r==-1) {
		return NULL ;
	}
	if(r<_IDX_INTSIZE) {
		return NULL ;
	}
	// Writing the index word
	r = write(db->filedes,idx->word,db->wlen) ;
	if(r==-1) {
		return NULL ;
	}
	if(r<db->wlen) {
		return NULL ;
	}
	r = write(db->filedes,"\n",1) ;
	if( (r==-1)||(r<1) ) {
		return NULL ;
	}
	// Writing record header: length and value
	r = _Idx_printint(idx->len,buf,_IDX_INTSIZE) ;
	if(r<_IDX_INTSIZE) {
		return NULL ;
	}
	buf[r] = ' ' ;
	r ++ ;
	r += snprintf(buf+r,_IDX_FLPSIZE+1,"%" _IDX_FLPFORM "e",idx->eval) ;
	if(r<_IDX_INTSIZE) {
		return NULL ;
	}
	buf[r] = ' ' ;
	r = lseek(db->filedes,idx->off+_IDX_HEADERLEN+_IDX_INTSIZE+1+db->size*(db->wlen+_IDX_INTSIZE+2),SEEK_SET) ;
	if(r==-1) {
		return NULL ;
	}
	r = write(db->filedes,buf,_IDX_INTSIZE+_IDX_FLPSIZE+2) ;
	if( (r==-1) || (r<(_IDX_INTSIZE+_IDX_FLPSIZE+2)) ) {
		return NULL ;
	}
	// Loop for writing record members
	i = 0 ;
	while(i<idx->len) {
		unsigned long j = 0 ;
		unsigned long k ;
		k = 0 ;
		while(k<db->flen) {
			rbuf[j] = idx->flds[i][k] ;
			j ++ ;
			k ++ ;
		}
		rbuf[j] = ':' ;
		j ++ ;
		k = 0 ;
		while(k<db->vlen) {
			rbuf[j] = idx->vals[i][k] ;
			j ++ ;
			k ++ ;
		}
		rbuf[j] = ' ' ;
		r = write(db->filedes,rbuf,db->flen+db->vlen+2) ;
		if( (r==-1)||(r<(db->flen+db->vlen+2)) ) {
			return NULL ;
		}
		i ++ ;
	}
	// Fix last character - should be a \n
	r = lseek(db->filedes,-1,SEEK_CUR) ;
	if(r==-1) {
		return NULL ;
	}
	r = write(db->filedes,"\n",1) ;
	if( (r==-1)||(r<1) ) {
		return NULL ;
	}
	return idx ;
}


#define _IDX_CHKMASK (1<<(8*sizeof(long)-1))


inline char _Idx_wordcmp(char * w1, char * w2, unsigned long s) {
	register unsigned long i = 0 ;
	while(i<s) {
		if(w1[i]<w2[i]) return -1 ;
		if(w1[i]>w2[i]) return +1 ;
		if(w1[i]==0) return 0 ;
		i ++ ;
	}
	return 0 ;
}

inline unsigned long _Idx_findword(Idx_DB db, char * word) {
	char * wbuf ;
	unsigned long i, k, s ;
	long r ;
	if(db==NULL) {
		return -1 ;
	}
	if(db->size==0) {
		return _IDX_CHKMASK ;
	}
	wbuf = alloca(db->wlen+1) ;
	if(wbuf==NULL) {
		return -1 ;
	}
	i = 0 ;
	s = db->size ;
	while(i<s) {
		k = (i+s)>>1 ;
		r = lseek(db->filedes,_IDX_HEADERLEN+_IDX_INTSIZE+1+k*(db->wlen+_IDX_INTSIZE+2),SEEK_SET) ;
		if(r==-1) {
			return -1 ;
		}
		r = read(db->filedes,wbuf,db->wlen) ;
		if(r==-1) {
			return -1 ;
		}
		if(r<db->wlen) {
			return -1 ;
		}
		switch(_Idx_wordcmp(word,wbuf,db->wlen)) {
			case -1:
				s=k ;
				break ;
			case +1:
				i=k+1 ;
				break ;
			default:
				return k&~_IDX_CHKMASK ;
		}
	}
	r = lseek(db->filedes,_IDX_HEADERLEN+i*(db->wlen+_IDX_INTSIZE+2),SEEK_SET) ;
	if(r==-1) {
		return -1 ;
	}
	r = read(db->filedes,wbuf,db->wlen) ;
	if(r==-1) {
		return -1 ;
	}
	if(r<db->wlen) {
		return -1 ;
	}
	if(_Idx_wordcmp(word,wbuf,db->wlen)==0) {
		return i&~_IDX_CHKMASK ;
	}
	return i|_IDX_CHKMASK ;
}

#define _IDX_MVBSZ 1024

inline unsigned long _Idx_moveblock(int fd, unsigned long beg, unsigned long end, long mov) {
	char buf[_IDX_MVBSZ] ;
	unsigned long l, i, r ;
	if(end<=beg) return 0 ;
	l = end - beg ;
	if(mov>0) {
//		fprintf(stderr,"  **  Moving block [%lu,%lu] -> %ld\n",beg,end,mov) ;
		i = l ;
		while(i>_IDX_MVBSZ) {
			i -= _IDX_MVBSZ ;
			r = lseek(fd,beg+i,SEEK_SET) ;
			if(r==-1) return -1 ;
			r = read(fd,buf,_IDX_MVBSZ) ;
			if( (r==-1)||(r<_IDX_MVBSZ) ) {
				return -1 ;
			}
			r = lseek(fd,beg+i+mov,SEEK_SET) ;
			if(r==-1) return -1 ;
			r = write(fd,buf,_IDX_MVBSZ) ;
			if( (r==-1)||(r<_IDX_MVBSZ) ) {
				return -1 ;
			}
		}
		if(i>0) {
			r = lseek(fd,beg,SEEK_SET) ;
			if(r==-1) return -1 ;
			r = read(fd,buf,i) ;
			if( (r==-1)||(r<i) ) {
				return -1 ;
			}
			r = lseek(fd,beg+mov,SEEK_SET) ;
			if(r==-1) return -1 ;
			r = write(fd,buf,i) ;
			if( (r==-1)||(r<i) ) {
				return -1 ;
			}
		}
	}
	else {
//		fprintf(stderr,"  **  Moving block [%lu,%lu] <- %ld\n",beg,end,-mov) ;
		i = 0 ;
		while((i+_IDX_MVBSZ)<l) {
			r = lseek(fd,beg+i,SEEK_SET) ;
			if(r==-1) return -1 ;
			r = read(fd,buf,_IDX_MVBSZ) ;
			if( (r==-1)||(r<_IDX_MVBSZ) ) {
				return -1 ;
			}
			r = lseek(fd,beg+i+mov,SEEK_SET) ;
			if(r==-1) return -1 ;
			r = write(fd,buf,_IDX_MVBSZ) ;
			if( (r==-1)||(r<_IDX_MVBSZ) ) {
				return -1 ;
			}
			i += _IDX_MVBSZ ;
		}
		if(i<l) {
			r = lseek(fd,beg+i,SEEK_SET) ;
			if(r==-1) return -1 ;
			r = read(fd,buf,l-i) ;
			if( (r==-1)||(r<(l-i)) ) {
				return -1 ;
			}
			r = lseek(fd,beg+i+mov,SEEK_SET) ;
			if(r==-1) return -1 ;
			r = write(fd,buf,l-i) ;
			if( (r==-1)||(r<(l-i)) ) {
				return -1 ;
			}
		}
	}
//	fprintf(stderr,"  **  Block moved, %lu chars\n",l) ;
	return l ;
}


Idx_Index Idx_IndexGet(Idx_DB db, char * word) {
	unsigned long i ;
	if(db==NULL) {
		return NULL ;
	}
	if(db->size==0) {
		return NULL ;
	}
	i = _Idx_findword(db,word) ;
	if(i==-1) {
		return NULL ;
	}
	if(i&_IDX_CHKMASK) {
		return NULL ;
	}
	i &= ~_IDX_CHKMASK ;
	return _Idx_IndexRead(db,i) ;
}


Idx_Index Idx_IndexAdd(Idx_DB db, char * word) {
	Idx_Index idx ;
	unsigned long i ;
	unsigned long r ;
	if(db==NULL) {
		return NULL ;
	}
	i = _Idx_findword(db,word) ;
	if(i&_IDX_CHKMASK) {
		// Create entry
		i &= ~_IDX_CHKMASK ;
		// Make room for table line
		r = _Idx_moveblock(db->filedes,
				_IDX_HEADERLEN+i*(db->wlen+_IDX_INTSIZE+2),
				_IDX_HEADERLEN+_IDX_INTSIZE+1+db->size*(db->wlen+_IDX_INTSIZE+2)+db->data,
				db->wlen+_IDX_INTSIZE+2) ;
		if(r==-1) {
			return NULL ;
		}
		// Allocate index structure and set members
		idx = Idx_IndexAlloc(db,word) ;
		if(idx==NULL) return NULL ;
		idx->pos = i ;
		idx->off = db->data ;
		// Increase data length and table size
		db->data += _IDX_INTSIZE+_IDX_FLPSIZE+2 ;
		db->size ++ ;
		if(_Idx_DBWrite(db)==NULL) {
			free(idx->word) ;
			free(idx) ;
			return NULL ;
		}
		// Write entry (table entry, and empty record at the end)
		if(_Idx_IndexWrite(db,idx,i)==NULL) {
			free(idx->word) ;
			free(idx) ;
			return NULL ;
		}
		return idx ;
	}
	// What do we do when there is already an existing entry?
	return _Idx_IndexRead(db,i) ;
}


Idx_Index Idx_IndexSet(Idx_DB db, Idx_Index idx) {
	char buf[_IDX_INTSIZE+1] ;
	Idx_Index tix ;
	unsigned long i ;
	unsigned long r, l, o ;
	if(db==NULL) {
		return NULL ;
	}
	if(idx==NULL) {
		return NULL ;
	}
	if(idx->db!=db) {
		return NULL ;
	}
	// Check for pos
	if( (idx->pos==-1)||(idx->pos>=db->size) ) {
		// Now we know registered indexes are no good
		idx->pos = -1 ;
		idx->off = -1 ;
	}
	else {
		tix = _Idx_IndexRead(db,idx->pos) ;
		if(_Idx_wordcmp(idx->word,tix->word,db->wlen)!=0) {
			idx->pos = -1 ;
			idx->off = -1 ;
		}
	}
	if(idx->pos==-1) {
		// Something was wrong with the index... find the right one
		i = _Idx_findword(db,idx->word) ;
		if(i==-1) return NULL ;
		if(i&_IDX_CHKMASK) {
			// Not found -> Create entry
			i &= ~_IDX_CHKMASK ;
			// Make room for table line
			r = _Idx_moveblock(db->filedes,
					_IDX_HEADERLEN+i*(db->wlen+_IDX_INTSIZE+2),
					_IDX_HEADERLEN+_IDX_INTSIZE+1+db->size*(db->wlen+_IDX_INTSIZE+2)+db->data,
					db->wlen+_IDX_INTSIZE+2) ;
			if(r==-1) {
				return NULL ;
			}
			// Set new members
			idx->pos = i ;
			idx->off = db->data ;
			// Increase data length and table size
			db->data += (db->flen+db->wlen+2)*idx->len+_IDX_INTSIZE+_IDX_FLPSIZE+2 ;
			db->size ++ ;
			if(_Idx_DBWrite(db)==NULL) {
				errno = EIO ;
				return NULL ;
			}
			// Write entry (table entry, and record at the end)
			if(_Idx_IndexWrite(db,idx,i)==NULL) {
				errno = EIO ;
				return NULL ;
			}
			return idx ;
		}
		// We have got the index, we go on
		idx->pos = i ;
	}
	// Doing real over-write
	tix = _Idx_IndexRead(db,idx->pos) ;
	if(tix==NULL) return NULL ;
	// Get exact record offset and original length
	idx->off = tix->off ;
	l = tix->len ;
	Idx_IndexFree(tix) ;
	// If we need to move things...
	if(idx->len!=l) {
		// Make/reclaim room for record length
		r = _Idx_moveblock(db->filedes,
				_IDX_HEADERLEN+_IDX_INTSIZE+1+db->size*(db->wlen+_IDX_INTSIZE+2)+idx->off+(db->flen+db->wlen+2)*l+_IDX_INTSIZE+_IDX_FLPSIZE+2,
				_IDX_HEADERLEN+_IDX_INTSIZE+1+db->size*(db->wlen+_IDX_INTSIZE+2)+db->data,
				(db->flen+db->wlen+2)*(idx->len-l)) ;
		// Correct data length
		if(idx->len>l) {
			// We increase the data segment size
			db->data += (db->flen+db->vlen+2)*(idx->len-l) ;
			if(_Idx_DBWrite(db)==NULL) {
				errno = EIO ;
				return NULL ;
			}
		}
		else {
			// We decrease the data segment size
			db->data -= (db->flen+db->vlen+2)*(l-idx->len) ;
			if(_Idx_DBWrite(db)==NULL) {
				errno = EIO ;
				return NULL ;
			}
			// We need to truncate file, as well
			ftruncate(db->filedes,_IDX_HEADERLEN+_IDX_INTSIZE+1+db->size*(db->wlen+_IDX_INTSIZE+2)+db->data) ;
		}
		// Correct offset values
		i = 0 ;
		while(i<db->size) {
			r = lseek(db->filedes,_IDX_HEADERLEN+i*(db->wlen+_IDX_INTSIZE+2),SEEK_SET) ;
			if(r==-1) {
				errno = EIO ;
				return NULL ;
			}
			r = read(db->filedes,buf,_IDX_INTSIZE+1) ;
			if((r==-1)||(r<(_IDX_INTSIZE+1))||(buf[_IDX_INTSIZE]!=' ')) {
				errno = EIO ;
				return NULL ;
			}
			if(_Idx_scanint(&o,buf,_IDX_INTSIZE)<_IDX_INTSIZE) {
				errno = EIO ;
				return NULL ;
			}
			if(o>idx->off) {
				// Offest is after the record we modified
				o += (db->flen+db->wlen+2)*(idx->len-l) ;
				r = _Idx_printint(o,buf,_IDX_INTSIZE) ;
				if(r<_IDX_INTSIZE) {
					errno = EIO ;
					return NULL ;
				}
				buf[_IDX_INTSIZE] = ' ' ;
				r = lseek(db->filedes,_IDX_HEADERLEN+i*(db->wlen+_IDX_INTSIZE+2),SEEK_SET) ;
				if(r==-1) {
					errno = EIO ;
					return NULL ;
				}
				r = write(db->filedes,buf,_IDX_INTSIZE+1) ;
				if(r==-1) {
					errno = EIO ;
					return NULL ;
				}
				if(r<_IDX_INTSIZE) {
					errno = EIO ;
					return NULL ;
				}
			}
			i ++ ;
		}
		// OK, everything has moved and is in order
	}
	// Write entry (table entry, and record at the end)
	if(_Idx_IndexWrite(db,idx,idx->pos)==NULL) {
		errno = EIO ;
		return NULL ;
	}
	return idx ;
}


void Idx_IndexDel(Idx_DB db, char * word) {
	char buf[_IDX_INTSIZE+1] ;
	Idx_Index tix ;
	unsigned long i, r, o, l ;
	if(db==NULL) {
		return ;
	}
	if(word==NULL) {
		return ;
	}
	// Find the right index
	i = _Idx_findword(db,word) ;
	// Not found... our job is already done
	if( (i==-1)||(i&_IDX_CHKMASK) ) return  ;
	// Read index from file
	tix = _Idx_IndexRead(db,i) ;
	if(tix==NULL) return ;
	// Get exact record offset
	o = tix->off ;
	// Get record length (in *bytes*)
	l = (db->flen+db->wlen+2)*tix->len+_IDX_INTSIZE+_IDX_FLPSIZE+2 ;
	Idx_IndexFree(tix) ;
	// Reclaim table line
	r = _Idx_moveblock(db->filedes,
			// beginning of next table line
			_IDX_HEADERLEN+(i+1)*(db->wlen+_IDX_INTSIZE+2),
			// beginning of old record segment
			_IDX_HEADERLEN+_IDX_INTSIZE+1+db->size*(db->wlen+_IDX_INTSIZE+2)+o,
			-(db->wlen+_IDX_INTSIZE+2)) ;
	// Move back end of record
	r = _Idx_moveblock(db->filedes,
			// end of old record segment
			_IDX_HEADERLEN+_IDX_INTSIZE+1+db->size*(db->wlen+_IDX_INTSIZE+2)+o+l,
			_IDX_HEADERLEN+_IDX_INTSIZE+1+db->size*(db->wlen+_IDX_INTSIZE+2)+db->data,
			-(db->wlen+_IDX_INTSIZE+2+l)) ;
	// Correct table size and data length
	db->data -= l ;
	db->size -- ;
	ftruncate(db->filedes,_IDX_HEADERLEN+_IDX_INTSIZE+1+db->size*(db->wlen+_IDX_INTSIZE+2)+db->data) ;
	if(_Idx_DBWrite(db)==NULL) {
		return ;
	}
	// Correct offset values
	i = 0 ;
	while(i<db->size) {
		r = lseek(db->filedes,_IDX_HEADERLEN+i*(db->wlen+_IDX_INTSIZE+2),SEEK_SET) ;
		if(r==-1) {
			return ;
		}
		r = read(db->filedes,buf,_IDX_INTSIZE+1) ;
		if((r==-1)||(r<(_IDX_INTSIZE+1))||(buf[_IDX_INTSIZE]!=' ')) {
			return ;
		}
		if(_Idx_scanint(&r,buf,_IDX_INTSIZE)<_IDX_INTSIZE) {
			return ;
		}
		if(r>o) {
			// Offest is after the record we modified
			r -= l ;
			r = _Idx_printint(r,buf,_IDX_INTSIZE) ;
			if(r<_IDX_INTSIZE) {
				return ;
			}
			buf[_IDX_INTSIZE] = ' ' ;
			r = lseek(db->filedes,_IDX_HEADERLEN+i*(db->wlen+_IDX_INTSIZE+2),SEEK_SET) ;
			if(r==-1) {
				return ;
			}
			r = write(db->filedes,buf,_IDX_INTSIZE+1) ;
			if(r==-1) {
				return ;
			}
			if(r<(_IDX_INTSIZE+1)) {
				return ;
			}
		}
		i ++ ;
	}
}


char * Idx_IndexWord(Idx_Index idx) {
	if(idx==NULL) return NULL ;
	return idx->word ;
}


double Idx_IndexEval(Idx_Index idx) {
	if(idx==NULL) return 0 ;
	return idx->eval ;
}


unsigned long Idx_IndexLen(Idx_Index idx) {
	if(idx==NULL) return -1 ;
	return idx->len ;
}



////////////////////////////////////////
// Index querying


void Idx_FieldAdd(Idx_Index idx, char * fld, char * val) {
	char * ttf ;
	char * ttv ;
	unsigned long i, j ;
	if(idx==NULL) {
		return ;
	}
	if(idx->db==NULL) {
		return ;
	}
	i = 0 ;
	while(i<idx->len) {
		j = 0 ;
		while(j<idx->db->flen) {
			if(fld[j]!=idx->flds[i][j]) break ;
			if(fld[j]==0) {
				j = 0 ;
				while(j<idx->db->vlen) {
					idx->vals[i][j] = val[j] ;
					if(val[j]==0) break ;
					j ++ ;
				}
				while(j<idx->db->vlen) {
					idx->vals[i][j] = 0 ;
					j ++ ;
				}
				return ;
			}
			j ++ ;
		}
		i ++ ;
	}
	if(idx->len==idx->mem) {
		char * * tf ;
		char * * tv ;
		tf = realloc(idx->flds,(idx->mem+_IDX_INDEX_BSIZE)*sizeof(unsigned long)) ;
		tv = realloc(idx->vals,(idx->mem+_IDX_INDEX_BSIZE)*sizeof(char *)) ;
		if((tf==NULL)||(tv==NULL)) return ;
		idx->flds = tf ;
		idx->vals = tv ;
		idx->mem += _IDX_INDEX_BSIZE ;
	}
	ttf = malloc((idx->db->flen+1)*sizeof(char)) ;
	if(ttf==NULL) {
		return ;
	}
	j = 0 ;
	while(j<idx->db->flen) {
		ttf[j] = fld[j] ;
		if(fld[j]==0) break ;
		j ++ ;
	}
	while(j<idx->db->flen) {
		ttf[j] = 0 ;
		j ++ ;
	}
	ttv = malloc((idx->db->flen+1)*sizeof(char)) ;
	if(ttv==NULL) {
		free(ttf) ;
		return ;
	}
	j = 0 ;
	while(j<idx->db->vlen) {
		ttv[j] = val[j] ;
		if(val[j]==0) break ;
		j ++ ;
	}
	while(j<idx->db->vlen) {
		ttv[j] = 0 ;
		j ++ ;
	}
	idx->flds[idx->len] = ttf ;
	idx->vals[idx->len] = ttv ;
	idx->len ++ ;
}


char * Idx_FieldGet(Idx_Index idx, char * fld) {
	unsigned long i, j ;
	if(idx==NULL) {
		return NULL ;
	}
	i = 0 ;
	while(i<idx->len) {
		j = 0 ;
		while(j<idx->db->flen) {
			if(fld[j]!=idx->flds[i][j]) break ;
			if(fld[j]==0) {
				return idx->vals[i] ;
			}
			j ++ ;
		}
		i ++ ;
	}
	return NULL ;
}


void Idx_FieldDel(Idx_Index idx, char * fld) {
	unsigned long i, j ;
	if(idx==NULL) {
		return ;
	}
	i = 0 ;
	while(i<idx->len) {
		j = 0 ;
		while(j<idx->db->flen) {
			if(fld[j]!=idx->flds[i][j]) break ;
			if(fld[j]==0) {
				idx->len -- ;
				free(idx->flds[i]) ;
				free(idx->vals[i]) ;
				idx->vals[i] = idx->vals[idx->len] ;
				idx->flds[i] = idx->flds[idx->len] ;
				if((idx->len+_IDX_INDEX_BSIZE)<idx->mem) {
					char * * tr ;
					char * * tv ;
					tr = realloc(idx->flds,(idx->mem-_IDX_INDEX_BSIZE)*sizeof(unsigned long)) ;
					tv = realloc(idx->vals,(idx->mem-_IDX_INDEX_BSIZE)*sizeof(char *)) ;
					if((tr==NULL)||(tv==NULL)) return ;
					idx->flds = tr ;
					idx->vals = tv ;
					idx->mem -= _IDX_INDEX_BSIZE ;
				}
				return ;
			}
			j ++ ;
		}
		i ++ ;
	}
}


char * Idx_FieldValByNum(Idx_Index idx, unsigned long num) {
	if(idx==NULL) {
		return NULL ;
	}
	if(num>=idx->len) {
		return NULL ;
	}
	return idx->vals[num] ;
}


char * Idx_FieldFldByNum(Idx_Index idx, unsigned long num) {
	if(idx==NULL) {
		return NULL ;
	}
	if(num>=idx->len) {
		return NULL ;
	}
	return idx->flds[num] ;
}





